import init_paths
import argparse
import yaml
import torch
from tqdm import tqdm
import numpy as np
from chainercv.evaluations import eval_detection_voc
from models import *
from torchvision.ops import nms
import utils.data_transforms as dt_trans
from datasets.pascal_voc import VOCDetection
from datasets.calipsa_pytorch_dataset import CalipsaDataset
from tasks.config import cfg, load_config
from utils.data_samplers import TestSampler, collate_minibatch, test_collate_minibatch
from layers.common_layers import bbox_regressor_transformation

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def evaluate(net, dataloader, test_count=-1, save_pred=False, iou_threshold=0.5):
    """Evaluates network."""
    if test_count == -1:
        test_count = len(dataloader)
    with torch.no_grad():
        net.eval()

        total_pred_boxes, total_pred_scores, total_pred_labels, total_gt_boxes, total_gt_labels = [], [], [], [], []

        cnt, identical_img_num = 0, len(cfg.TEST.BBOX_AUG.SCALES) * (cfg.TEST.BBOX_AUG.H_FLIP * 2)
        for batch_imgs, batch_target, batch_boxes, ori_batch_boxes, batch_gt_labels, batch_gt_boxes in tqdm(dataloader,
                                                                                                         "Evaluation"):
            cnt += 1

            if cnt > test_count:
                break

            if batch_gt_labels[0] is None:
                continue

            batch_imgs, batch_target = (batch_imgs.to(DEVICE), batch_target.to(DEVICE))
            batch_boxes = batch_boxes.to(DEVICE)
            batch_gt_boxes[0] = batch_gt_boxes[0].to(DEVICE)

            net_outputs = net(batch_imgs, batch_boxes, batch_target)

            if cnt % identical_img_num == 1:
                combined_scores = net_outputs['final_scores'][:, 1:]
                continue
            elif cnt % identical_img_num == 0:
                combined_scores += net_outputs['final_scores'][:, 1:]
                combined_scores /= identical_img_num
            else:
                combined_scores += net_outputs['final_scores'][:, 1:]
                continue

            proposal_boxes = ori_batch_boxes.to(DEVICE)
            if 'regression_boxes' in net_outputs:
                pred_offsets = net_outputs['regression_offsets']
                proposal_boxes = bbox_regressor_transformation(proposal_boxes, pred_offsets)
            num_classes = combined_scores.shape[1]

            batch_pred_boxes, batch_pred_scores, batch_pred_labels = [], [], []

            for i in range(num_classes):
                region_scores = combined_scores[:, i]
                score_mask = region_scores > 0.005

                selected_scores = region_scores[score_mask]
                selected_boxes = proposal_boxes[score_mask]

                nms_mask = nms(selected_boxes, selected_scores, 0.4)
                batch_pred_boxes.append(selected_boxes[nms_mask].cpu().numpy())
                batch_pred_scores.append(selected_scores[nms_mask].cpu().numpy())
                batch_pred_labels.append(np.full(len(nms_mask), i, dtype=np.int32))

            np_batch_pred_boxes = np.concatenate(batch_pred_boxes, axis=0)
            np_batch_pred_labels = np.concatenate(batch_pred_labels, axis=0)
            np_batch_pred_scores = np.concatenate(batch_pred_scores, axis=0)
            if test_count <= 20:
                print('cnt: \n')
                print('pred_boxes:\n', np_batch_pred_boxes)
                print('pred_scores:\n', np_batch_pred_scores)
                print('pred_labels:\n', np_batch_pred_labels)
                print('gt_boxes:\n', batch_gt_boxes[0].cpu().numpy())
                print('gt_labels:\n', batch_gt_labels[0].cpu().numpy())

            if save_pred:
                np.save(f'pred/pred_boxes_{cnt}.npy', np_batch_pred_boxes)
                np.save(f'pred/pred_labels_{cnt}.npy', np_batch_pred_labels)
                np.save(f'pred/pred_scores_{cnt}.npy', np_batch_pred_scores)
                np.save(f'pred/pred_conf_map_{cnt}.npy', net_outputs['conf_map'].cpu().numpy())

            total_pred_boxes.append(np_batch_pred_boxes)
            total_pred_scores.append(np_batch_pred_scores)
            total_pred_labels.append(np_batch_pred_labels)
            total_gt_boxes.append(batch_gt_boxes[0].cpu().numpy())
            total_gt_labels.append(batch_gt_labels[0].cpu().numpy())

        result = eval_detection_voc(total_pred_boxes, total_pred_labels, total_pred_scores,
                                    total_gt_boxes, total_gt_labels, iou_thresh=iou_threshold, use_07_metric=True)

        tqdm.write(f"Avg AP: {result['ap']} (iou_threshold: {iou_threshold})")
        tqdm.write(f"Avg mAP: {result['map']} (iou_threshold: {iou_threshold})")

        result_2 = eval_detection_voc(total_pred_boxes, total_pred_labels, total_pred_scores,
                                    total_gt_boxes, total_gt_labels, iou_thresh=0.01, use_07_metric=True)

        tqdm.write(f"Avg AP: {result_2['ap']} (iou_threshold: 0.01)")
        tqdm.write(f"Avg mAP: {result_2['map']} (iou_threshold: 0.01)")

        net.train()
        return result['map']


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Evaluate trained model")
    parser.add_argument("--state_path", help="Path of trained model's state")
    parser.add_argument("--test_count", type=int, default=-1, help="Batch Size")
    parser.add_argument("--min_size", type=int, default=35, help="minimum size for filtering small boxes")
    parser.add_argument("--iou_threshold", type=float, default=0.5, help="iou threshold")
    parser.add_argument("--save_pred", type=bool, default=False, help="save_pred")
    parser.add_argument("--dataset", type=str, default="calipsa", help="Dataset to use")
    parser.add_argument("--roi_head", type=str, default="oicr", help="Roi head to use")
    parser.add_argument("--base_net", type=str, default="vgg", help="Base network to use")
    parser.add_argument("--debug", type=bool, default=False, help="DEBUG")
    parser.add_argument('--cfg', dest='cfg_file', default="configs/baselines/vgg16_voc2007.yaml",
                        help='Config file for training (and optionally testing)')
    parser.add_argument('--model', help='Set model', default="regressor_oicr", type=str)
    args = parser.parse_args()
    load_config(args.cfg_file)

    transforms = dt_trans.Compose([
        dt_trans.Normalize([102.9801, 115.9465, 122.7717]),
        dt_trans.HorizontalFlip(),
        dt_trans.Resize(),
        dt_trans.ToTensor(),
    ])

    if args.dataset == 'voc_2007':
        dataset = VOCDetection(cfg.DATA_DIR + '/', year='2007', image_set='test', transforms=transforms)
    elif args.dataset == 'calipsa':
        dataset = CalipsaDataset('test', transforms=transforms, min_size=args.min_size)
    batchSampler = TestSampler(len(dataset))

    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_sampler=batchSampler,
        num_workers=cfg.DATA_LOADER.NUM_THREADS,
        collate_fn=test_collate_minibatch)

    num_classes, input_channels = 20, 3
    if args.dataset == 'calipsa':
        num_classes, input_channels = 4, 6
    net = eval(args.model).loot_model(
        args, num_classes=num_classes, input_channels=input_channels, backbone='vgg16')
    if not args.debug:
        checkpoint = torch.load(args.state_path, map_location=lambda storage, loc: storage)
        net.load_state_dict(checkpoint['model'], strict=True)
    net.to(DEVICE)

    evaluate(net, dataloader, test_count=args.test_count, save_pred=args.save_pred, iou_threshold=args.iou_threshold)
