# train detection model
import init_paths
import argparse
import os
import sys
import pickle
import traceback
import random
import numpy as np
import yaml
import torch
from models import *
from datasets.pascal_voc import VOCDetection
from datasets.calipsa_pytorch_dataset import CalipsaDataset
from tasks.config import cfg, load_config
from tasks.evaluate import evaluate
import utils.optimizer_utils as optimizer_utils
import utils.misc as misc_utils
import utils.data_transforms as dt_trans
from utils.loss_printer import LossPrinter

from utils.data_samplers import TrainSampler, TestSampler, collate_minibatch, test_collate_minibatch
from utils.detectron_weight_helper import load_detectron_weight

torch.manual_seed(10)
torch.cuda.manual_seed_all(10)
np.random.seed(10)
random.seed(10)
torch.backends.cudnn.deterministic = True

torch.multiprocessing.set_sharing_strategy('file_system')


def parse_args():
    """Parse input arguments"""
    parser = argparse.ArgumentParser(description='Train a WSOD detection network')

    parser.add_argument('--cfg', dest='cfg_file', default="configs/baselines/vgg16_voc2007.yaml",
                        help='Config file for training (and optionally testing)')
    parser.add_argument('--backbone', dest='backbone', default="vgg16",
                        help='Backbone of Model')
    # Optimization
    # These options has the highest prioity and can overwrite the values in config file
    # or values set by set_cfgs. `None` means do not overwrite.
    parser.add_argument('--bs', dest='batch_size',
                        help='Explicitly specify to overwrite the value come from cfg_file.', type=int, default=1)

    parser.add_argument('--nw', dest='num_workers',
                        help='Explicitly specify to overwrite number of workers to load data. Defaults to 3', type=int,
                        default=3)
    parser.add_argument('--iter_size', help='Update once every iter_size steps, as in Caffe.', type=int, default=4)

    parser.add_argument('--o', dest='optimizer', help='Training optimizer.', default='SGD')
    parser.add_argument('--lr', help='Base learning rate.', default=5e-5, type=float)

    # Epoch
    parser.add_argument('--start_step', help='Starting step count for training epoch. 0-indexed.', default=0, type=int)
    parser.add_argument('--end_step', help='total steps.', default=40000, type=int)

    # Resume training: requires same iterations per epoch
    parser.add_argument('--no_save', help='do not save anything', action='store_true')
    parser.add_argument('--model', help='Set model', default="regressor_oicr", type=str)
    parser.add_argument('--dataset', help='dataset', default="voc_2007", type=str)
    parser.add_argument('--state_path', help='load state', default=None, type=str)
    return parser.parse_args()


def save_ckpt(output_dir, args, step, train_size, model, optimizer, is_bkp=False):
    """Save checkpoint"""
    if args.no_save:
        return
    ckpt_dir = os.path.join(output_dir, 'ckpt')
    if not os.path.exists(ckpt_dir):
        os.makedirs(ckpt_dir)

    if is_bkp:
        save_name = os.path.join(ckpt_dir, 'bkp.pth')
    else:
        save_name = os.path.join(ckpt_dir, 'model_step{}.pth'.format(step))

    torch.save({
        'step': step,
        'train_size': train_size,
        'batch_size': args.batch_size,
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict()}, save_name)


def main():
    """Main function"""

    if not torch.cuda.is_available():
        sys.exit("Need a CUDA device to run the code.")
    cfg.CUDA = True

    args = parse_args()
    print('Called with args:')
    print(args)

    load_config(args.cfg_file)
    print('Number of data loading threads: %d' % args.num_workers)

    transforms = dt_trans.Compose([
        dt_trans.Normalize([102.9801, 115.9465, 122.7717]),
        dt_trans.HorizontalFlip(),
        dt_trans.Resize(),
        dt_trans.ToTensor(),
    ])

    if args.dataset == 'voc_2007':
        dataset = VOCDetection(cfg.DATA_DIR + '/', year='2007', image_set='trainval', transforms=transforms)
        test_dataset = VOCDetection(cfg.DATA_DIR + '/', year='2007', image_set='test', transforms=transforms)
    elif args.dataset == 'calipsa':
        dataset = CalipsaDataset('train', transforms=transforms)
        test_dataset = CalipsaDataset('test', transforms=transforms)

    train_size = len(dataset) // args.batch_size * args.batch_size

    batchSampler = TrainSampler(
        subdivision=args.iter_size,
        batch_size=args.iter_size,
        max_iterations=args.end_step,
        num_samples=len(dataset),
        image_scales=cfg.TRAIN.SCALES,
        scale_interval=10)
    testBatchSampler = TestSampler(len(test_dataset))

    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_sampler=batchSampler,
        num_workers=args.num_workers,
        collate_fn=collate_minibatch)
    test_dataloader = torch.utils.data.DataLoader(
        test_dataset,
        batch_sampler=testBatchSampler,
        num_workers=args.num_workers,
        collate_fn=test_collate_minibatch)
    dataiterator = iter(dataloader)

    # Model
    num_classes, input_channels = 20, 3
    if args.dataset == 'calipsa':
        num_classes, input_channels = 4, 6
    model = eval(args.model).loot_model(
        args, num_classes=num_classes, input_channels=input_channels, backbone=args.backbone)

    if cfg.CUDA:
        model.cuda()

    if args.state_path is not None:
        checkpoint = torch.load(args.state_path, map_location=lambda storage, loc: storage)
        model.load_state_dict(checkpoint['model'], strict=True)

    # Optimizer
    bias_params, bias_param_names, nonbias_params, nonbias_param_names, nograd_param_names = [], [], [], [], []
    for key, value in model.named_parameters():
        if value.requires_grad:
            if 'bias' in key:
                bias_params.append(value)
                bias_param_names.append(key)
            else:
                nonbias_params.append(value)
                nonbias_param_names.append(key)
        else:
            nograd_param_names.append(key)

    params = [
        {'params': nonbias_params, 'lr': args.lr, 'weight_decay': cfg.SOLVER.WEIGHT_DECAY},
        {'params': bias_params, 'lr': args.lr * (cfg.SOLVER.BIAS_DOUBLE_LR + 1),
         'weight_decay': cfg.SOLVER.WEIGHT_DECAY if cfg.SOLVER.BIAS_WEIGHT_DECAY else 0},
    ]

    optimizer = None
    if args.optimizer == "SGD":
        optimizer = torch.optim.SGD(params, momentum=cfg.SOLVER.MOMENTUM)
    elif args.optimizer == "Adam":
        optimizer = torch.optim.Adam(params)
    scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[20000, 40000], gamma=0.1)

    # Training Setups
    args.run_name = misc_utils.get_run_name()
    output_dir = misc_utils.get_output_dir(args)
    args.cfg_filename = os.path.basename(args.cfg_file)

    if not args.no_save:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        blob = {'cfg': yaml.dump(cfg), 'args': args}
        with open(os.path.join(output_dir, 'config_and_args.pkl'), 'wb') as f:
            pickle.dump(blob, f, pickle.HIGHEST_PROTOCOL)

    # Training Loop
    model.train()
    loss_several_steps, best_test_ap = [], 0.0
    loss_printer = LossPrinter(args.run_name, args.end_step)
    try:
        print('Training starts !')
        for step in range(args.start_step, args.end_step):

            optimizer.zero_grad()
            inner_total_loss_list, inner_iter = [], 0
            while inner_iter < args.iter_size:
                try:
                    input_data = next(dataiterator)
                except StopIteration:
                    dataiterator = iter(dataloader)
                    input_data = next(dataiterator)

                im_data = input_data[0].cuda()
                labels = input_data[1].cuda()
                rois = input_data[2].cuda()
                if labels.sum() == 0:
                    continue
                inner_iter += 1
                model.set_inner_iter(inner_iter)

                net_outputs = model(im_data, rois, labels)
                total_loss = 0
                for k, loss in net_outputs['losses'].items():
                    assert loss.shape[0] == cfg.NUM_GPUS
                    loss = loss.mean(dim=0, keepdim=True)
                    total_loss += loss

                total_loss.backward(retain_graph=True)
                inner_total_loss_list.append(total_loss.item())
                loss_printer.update(net_outputs)
            loss_several_steps.append(np.mean(inner_total_loss_list))
            optimizer.step()
            scheduler.step()

            if step % 20 == 0 or step == args.end_step - 1:
                loss_printer.print_loss_status(step, loss_several_steps)
                loss_several_steps = []

            if (step + 1) % 1000 == 0 and args.model != 'vgg16_classification':
                period_ap = evaluate(model, test_dataloader, test_count=100, save_pred=False, iou_threshold=0.5)
                if period_ap > best_test_ap:
                    best_test_ap = period_ap
                    if step >= 4999:
                        save_ckpt(output_dir, args, step, train_size, model, optimizer)

            if args.model == 'vgg16_classification' and step >= 4999 and (step + 1) % 1000 == 0:
                save_ckpt(output_dir, args, step, train_size, model, optimizer)

    except (RuntimeError, KeyboardInterrupt):
        del dataiterator
        stack_trace = traceback.format_exc()
        print(stack_trace)


if __name__ == '__main__':
    main()
